from typing import List, Dict
import pandas as pd

from rainiee_lib.objects.orders import order


class HoldingEntries(object):
    def construct(self,symbol :str, holding_qty : int, holding_price : float):
        self.symbol = symbol
        # 当前持仓数量
        self.holding_qty = 0
        # 当前持仓成本
        self.holding_price = 0
        # 总投入金额
        self.total_amt = 0
        # 浮动盈亏
        self.profit_loss_amt = 0

    def get_holding_return(self, curr_price) -> float:
        if self.holding_qty == 0:
            return self.profit_loss_amt
        return self.holding_qty * (curr_price - self.holding_price) + self.profit_loss_amt

    def place_order(self, order_obj : order.Order):
        if order_obj.buy_order:#买入
            total_qty = self.holding_qty + order_obj.order_qty
            self.holding_price = (self.holding_price * self.holding_qty + order_obj.get_filled_amount()) / total_qty
            self.holding_qty = total_qty
            self.total_amt = self.total_amt + order_obj.get_filled_amount()
        else:#卖出
            if (self.holding_qty < order_obj.order_qty):
                print("报错，卖出数量应小于持有数量")
                return None
            else:
                total_qty = self.holding_qty - order_obj.order_qty
                self.holding_price =  (self.holding_price * self.holding_qty - order_obj.get_filled_amount())/total_qty
                self.holding_qty = self.holding_qty - order_obj.order_qty
                self.profit_loss_amt = self.profit_loss_amt + order_obj.get_filled_amount()

class PortfolioHolding(object):
    #当有unique key的时候需要用map进行key的索引
    def construct(self, portf_holdings_map : Dict[str,HoldingEntries]):
        self.portf_holdings_map = portf_holdings_map
        self.portf_holdings_profit_map = {}

    def construct_from_list(self, portf_holdings : List[HoldingEntries]):
        self.portf_holdings_map = {x.symbol: x for x in portf_holdings}

    def rebalance(self, order_obj : order.Order):
        if order_obj.symbol in self.portf_holdings_map.keys():
            self.get_holding_entry(order_obj.symbol).place_order(order_obj)
        else:
            holding_entry = HoldingEntries()
            holding_entry.construct(order_obj.symbol, 0, 0)
            holding_entry.place_order(order_obj)
            self.portf_holdings_map.update({holding_entry.symbol:holding_entry})

    def get_holding_entry(self, symbol : str) -> HoldingEntries:
        return self.portf_holdings_map.get(symbol)


    def get_all_holding_entry(self) -> dict:
        return self.portf_holdings_map

    def get_portf_holding_df(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.portf_holdings_map,orient="index")

 



import uuid
from typing import List

from rainiee_lib.objects import objects_utils
from rainiee_lib.objects.backtesting.backtest_result import BacktestResult
from rainiee_lib.objects.base.base_object import BaseObject


class BacktestStrategyHistory(BaseObject):

    def construct(self,backtest_result:BacktestResult,bt_result_sell_info:{},is_sell):
        # 每次组合的id标识
        self.uuid = 'HIS' + uuid.uuid1().hex
        # 当前组合的回测结果
        self.backtest_result = backtest_result
        # 是否卖出
        self.is_sell = is_sell
        # 给当前组合的每一天绑定信号
        return_list = backtest_result.return_list
        for _return in return_list:
            objects_utils.set_value(_return,'selling_signal_info',bt_result_sell_info.get(objects_utils.get_value(_return,'index')))
        return self



class BacktestStrategyResult(BaseObject):

    def construct(self,
                  backtest_history_list:List[BacktestStrategyHistory],
                  benchmark_symbol_list
                  ):
        # 策略回测历史记录
        self.backtest_history_list = backtest_history_list
        # 策略滚动收益
        self.p_return_roll_total = 1

        # 策略滚动基准收益相关，格式
        '''
        {
          p_return_roll_total_399300.SZ:1.150001
        }
        '''
        self.calculate(benchmark_symbol_list)
        return self


    def calculate(self,benchmark_symbol_list):
        for benchmark_symbol in benchmark_symbol_list:
            setattr(self,'p_return_roll_total_' + benchmark_symbol,1)

        # 计算history滚动收益
        for strategy_history in self.backtest_history_list:
            self.p_return_roll_total = self.p_return_roll_total * strategy_history.backtest_result.p_return_total
            for benchmark_symbol in benchmark_symbol_list:
                benchmark_key = 'p_return_roll_total_' + benchmark_symbol
                setattr(self,benchmark_key,(
                        getattr(self,benchmark_key) * (objects_utils.get_value(strategy_history.backtest_result,'p_return_total_' + benchmark_symbol))
                ))




from rainiee_lib.dataframe import pandas_df
from rainiee_lib.constants import data_consts
from rainiee_lib.objects import objects_utils
import pandas as pd
import json
from typing import List

class PortfolioEntries(object):
    def construct(self, symbol : str, weights : float, est_ret : float, ext = {}):
        self.symbol = symbol
        self.weights = round(float(weights), 6)
        self.est_ret = round(float(est_ret), 6)
        self.ext = ext

    def to_json(self):
        return objects_utils.get_fields_json(self, objects_utils.portfolio_defualt_fields)

    def from_json(self, json_input):
        self.construct(json_input['symbol'],json_input['weights'],json_input['est_ret'],json_input['ext'])
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))

class PortfolioParams(object):
    def construct(self, mean_returns : pd.DataFrame, cov_matrix : pd.DataFrame, ext={}):
        self.mean_returns = mean_returns
        self.cov_matrix = cov_matrix
        self.ext = ext

    def to_json(self):
        return {
            "mean_returns" : pandas_df.json_write_dataframe(self.mean_returns),
            "cov_matrix" : pandas_df.json_write_dataframe(self.cov_matrix),
            "ext" : self.ext
                }
    def from_json(self,json_input):
        self.construct(pandas_df.json_read_dataframe(json_input['mean_returns']),
                       pandas_df.json_read_dataframe(json_input['cov_matrix']),
                       json_input['ext'])
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))


class Portfolio(object):
    def construct(self,
                  portf_allocation : List[PortfolioEntries],
                  est_portf_return : float,
                  est_portf_var : float,
                  solver_info : dict,
                  portf_param : PortfolioParams,
                  portf_model : str,
                  ext = {}):
        self.portf_allocation = portf_allocation
        self.est_portf_return = round(float(est_portf_return),6)
        self.est_portf_var = round(float(est_portf_var),6)
        self.solver_info = solver_info
        self.portf_param = portf_param
        self.portf_model = portf_model
        self.ext = ext

    def est_portf_sharpe(self):
        return round(self.est_portf_return / self.est_portf_var ,6)

    def get_portf_df(self):
        return pandas_df.get_df_obj_list(self.portf_allocation).sort_values(by=[data_consts.WEIGHTS], ascending=False)

    def get_portf_list(self):
        return list(self.get_portf_df().T.to_dict().values())

    def get_top_k_list(self, k):
        res_df = self.get_portf_df().sort_values(by=[data_consts.WEIGHTS], ascending=False)[:int(k)]
        res_df[data_consts.WEIGHTS] = res_df[data_consts.WEIGHTS] / res_df[data_consts.WEIGHTS].sum()
        res_list = list(res_df.T.to_dict().values())
        return res_list

    def to_json(self):
        return {"portf_allocation": [o.to_json() for o in self.portf_allocation],
                "est_portf_return": self.est_portf_return, "est_portf_var": self.est_portf_var,
                "solver_info": self.solver_info, "portf_param": self.portf_param.to_json(),
                "portf_model":self.portf_model, "ext": self.ext}
    def from_json(self, json_input):
        self.construct(objects_utils.from_json_object_list("PortfolioEntries", json_input['portf_allocation']),
                       json_input['est_portf_return'],
                       json_input['est_portf_var'],
                       json_input['solver_info'],
                       objects_utils.from_json_object("PortfolioParams", json_input['portf_param']),
                       json_input['portf_model'],
                       json_input['ext'])
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))





import pandas as pd
from typing import List

def json_write_dataframe_list(dataframe_list : List[pd.DataFrame]):
    json_dataframe_list = []
    for dataframe in dataframe_list:
        json_dataframe_list.append(json_write_dataframe(dataframe))
    return json_dataframe_list

def json_read_dataframe_list(json_dataframe_list) -> List[pd.DataFrame]:
    dataframe_list = []
    for json_dataframe in json_dataframe_list:
        dataframe_list.append(json_read_dataframe(json_dataframe))
    return dataframe_list

def json_write_dataframe(dataframe : pd.DataFrame):
    return dataframe.to_json(orient='table')

def json_read_dataframe(json_dataframe) -> pd.DataFrame:
    return pd.read_json(json_dataframe, orient='table')

#read in list of objects and returns pandas dataframe
def get_df_obj_list(list_obj):
    return pd.DataFrame.from_records([o.__dict__ for o in list_obj])


from matplotlib import pyplot as plt
from ..lib import utils
class Plot(object):
    def __init__(self):
        utils.info('Initializing Plotting Object')

    """
    输入横轴行为column数据（例如股票symbol）纵轴为时间序列（例如日期index）返回plt对象
            策略A       上证指数
    7500    1.00112    1.00312 
    7501    1.00212    1.00412 
    
    """
    def plot_timeseries_data(self, dataframe):
        (dataframe / dataframe.iloc[0] * 100).plot(figsize=(10, 5))
        return plt



import json

from rainiee_lib.objects.portfolio import portfolio
from rainiee_lib.objects.stocks import stocks, indexs
from typing import List

portfolio_defualt_fields = ["symbol","weights","est_ret","ext"]
def get_fields_json(obj : object, fields : List[str]) -> dict:
    json_output = {}
    for field in fields:
        json_output.update({field: getattr(obj, field)})
    return json_output

def to_json_object_list(object_list : list):
    return [obj.to_json() for obj in object_list]

def from_json_object_list(object_type :str, json_list: List[dict]):
    append_list = []
    for input in json_list:
        instance = get_instance(object_type)
        instance.from_json(input)
        append_list.append(instance)
    return append_list

def from_json_object(object_type: str, json_input : dict):
    instance = get_instance(object_type)
    instance.from_json(json_input)
    return instance

def get_instance(object_type:str):
    if (object_type == "Stock"):
        return stocks.Stock()
    if (object_type == "StockHistory"):
        return stocks.StockHistory()
    if (object_type == "Index"):
        return indexs.Index()
    if (object_type == "IndexHistory"):
        return indexs.IndexHistory()
    if (object_type == "StockHistoryFetcher"):
        return stocks.StockHistoryFetcher()
    if (object_type == "PortfolioEntries"):
        return portfolio.PortfolioEntries()
    if (object_type == "PortfolioParams"):
        return portfolio.PortfolioParams()
    return None

def get_value(obj:object,key:str):
    if isinstance(obj,dict):
        return obj.get(key)
    return getattr(obj,key)

def set_value(self:object,key:str,value:object):
    if isinstance(self,dict):
        self[key] = value
        return
    setattr(self,key,value)


def set_attrs(obj:object,input_dict:dict):
    for key,value in input_dict.items():
        setattr(obj,key,value)


#数据常量
INDEX = 'index'
DATE = 'date'
SYMBOL = 'symbol'
PRE_CLOSE_PRICE = 'pre_close_price'
CLOSE_PRICE = 'close_price'
OPEN_PRICE = 'open_price'
HIGH_PRICE = 'high_price'
LOW_PRICE = 'low_price'
TURNOVER_RATE = 'turnover_rate'
VOLUME = 'volume'
AMOUNT = 'amount'
TOTAL_MKT_VAL = 'total_market_value'
CIRC_MKT_VAL = 'circulate_market_value'
MEAN_RETURNS = 'return'
WEIGHTS = 'weights'

# BACKTEST
DAILY_P_RETURN = 'daily_p_return'
DAILY_P_RETURN_TOTAL = 'daily_p_return_total'
AVG_PROFIT = 'avg_profit'
GEOMETRIC_MEAN = 'geometric_mean'
MIN_HOLD_PROFIT = 'min_hold_profit'
MAX_HOLD_PROFIT = 'max_hold_profit'
MIN_HOLD_PROFIT_INDEX = 'min_hold_profit_index'
MAX_HOLD_PROFIT_INDEX = 'max_hold_profit_index'


def build_field_sql(field_names):
    field_str = ''
    for column in field_names:
        field_str = field_str + column + ','
    return field_str[:-1]


def build_end_sql(uq_field_names, field_names):
    uq_field_str = ''
    for uq_field_name in uq_field_names:
        uq_field_str = uq_field_str + uq_field_name + ','

    end_sql = ' on conflict (' + (uq_field_str[:-1]) + ') do update set '
    for column in field_names:
        end_sql = end_sql + (column + '=' + 'excluded.' + column) + ','
    return end_sql[:-1]


def batch_insert(table_name, field_names, uq_field_names, list_data):
    sql_head = 'INSERT INTO ' + table_name + ' (' + build_field_sql(field_names) + ') VALUES '
    sql_items = ''
    for row in list_data:
        sql_items = sql_items + '( '
        for column in field_names:
            val = row[column]
            if isinstance(val, str):
                val = "'" + val + "'"
            else:
                val = str(val)
            sql_items = sql_items + val + ','
        sql_items = sql_items[:-1] + '),'

    sql_items = sql_items[:-1]
    sql_end = build_end_sql(uq_field_names, field_names)
    sql = sql_head + sql_items + sql_end
    from django.db import connection
    cursor = connection.cursor()
    cursor.execute(sql, None)



import json
from typing import List

from rainiee_lib.constants.data_consts import *
from rainiee_lib.objects import objects_utils
from rainiee_lib.objects.base.base_object import BaseObject
from rainiee_lib.objects.portfolio.portfolio import PortfolioEntries
from rainiee_lib.objects.stocks.indexs import IndexHistoryFetcher
from rainiee_lib.objects.stocks.stocks import Stock, StockHistoryFetcher
import pandas as pd

class PortfolioReturn(BaseObject):

    def construct(self,
                  portfolios:List[PortfolioEntries],
                  items:List[Stock],
                  index:int,
                  daily_p_return:float,
                  daily_p_return_total:float,
                  geometric_mean:float,
                  avg_profit:float,
                  min_hold_profit:float,
                  min_hold_profit_index:int,
                  max_hold_profit:float,
                  max_hold_profit_index:int,
                  benchmark_profit:dict
                  ):
        # 策略组合
        self.portfolios = portfolios
        # 交易日
        self.index = index
        # 策略组合交易日内，个股收益明细
        self.items = items
        # 组合当日收益
        self.daily_p_return = daily_p_return
        # 组合滚动收益
        self.daily_p_return_total = daily_p_return_total
        # 滚动几何平均数
        self.geometric_mean = geometric_mean
        # 算数平均数
        self.avg_profit = avg_profit
        # 滚动最小收益
        self.min_hold_profit = min_hold_profit
        # 滚动最小收益对应的交易日
        self.min_hold_profit_index = min_hold_profit_index
        # 滚动最大收益
        self.max_hold_profit = max_hold_profit
        # 滚动最大收益对应的交易日
        self.max_hold_profit_index = max_hold_profit_index

        # 当天交易日基准收益相关，格式
        '''
        {
          daily_p_return_399300.SZ:1.020001,
          daily_p_return_total_399300.SZ:1.150001
        }
        '''
        objects_utils.set_attrs(self,benchmark_profit)
        return self

    def to_json_str(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def from_json(self,json_data):
        self.__dict__ = json_data


class BacktestResult(BaseObject):

    def construct(self,
                  return_list:List[PortfolioReturn],
                  benchmark_symbol_list:list
                  ):
        last_return = return_list[len(return_list)-1]
        # 回测历史记录
        self.return_list = return_list
        # 回测总收益
        self.p_return_total = objects_utils.get_value(last_return,'daily_p_return_total')
        # 买入日
        self.buy_index = objects_utils.get_value(return_list[0],'index')
        # 持仓天数
        self.hold_day = objects_utils.get_value(last_return,'index') -  self.buy_index + 1

        #回测总收益基准相关，格式
        '''
        {
          p_return_total_399300.SZ:1.150001
        }
        '''
        for benchmark_symbol in benchmark_symbol_list:
            field_name = 'p_return_total_'+benchmark_symbol
            setattr(self,field_name,objects_utils.get_value(last_return,'daily_p_return_total_'+benchmark_symbol))
        return self



import json

import requests

from rainiee_lib.lib.utils import *



class FeishuClient(object):
    def __init__(self):
        info('feishu client init')
        self.headers = {'Content-Type': 'application/json'}

    def send(self,message,url):
        req_json = {
            "msg_type":"text",
            "content":{"text":message}
        }
        res = requests.post(url,json = req_json ,headers=self.headers)
        info(res.text)

    def mass(self,message,urls):
        for url in urls:
            self.send(message,url)


#!/usr/bin/python
# -*- coding: UTF-8 -*-
import time
import traceback
import logging
import uuid
from datetime import datetime

logger = logging.getLogger('logger')
import threading
local_data = threading.local()

def set_thread_local_trac_id(func_name):
    local_data.trac_id = uuid.uuid1().hex
    return local_data.trac_id

def remove_thread_local_trac_id():
    return True


def add_all_dic(origin_dic, source_dic):
    for key in source_dic:
        origin_dic[key] = source_dic[key]
    return origin_dic

def info(message):
    try:
        logger.info(message)
    except Exception as e:
        print(e)


def warn(message):
    try:
        logger.warning(message)
    except Exception as e:
        print(e)


def error(message):
    try:
        logger.error(message)
    except Exception as e:
        print(e)



def list_to_dic(items, keys):
    if items is None:
        items = []
    if not isinstance(items, list):
        items = [items]
    res_dic = {}
    for item in items:
        dic_key = ''
        for key in keys:
            dic_key = dic_key + str(item[key]) + '_'
        res_dic[dic_key[:-1]] = item
    return res_dic

def get_data(retry_max_count, func, *args):
    data = None
    try:
        if retry_max_count > 0:
            data = func(*args)
    except Exception as e:
        error(traceback.format_exc())
        retry_max_count = retry_max_count - 1
        time.sleep(0.3)
        return get_data(retry_max_count, func, *args)
    return data

def check_trade_range_time(start_time_interval,end_time_interval,rainiee_client):
    from interval import Interval
    # 先判断当前时间是否在时间区间内
    # 当前时间
    now_localtime = time.strftime("%H:%M:%S", time.localtime())
    # 当前时间（以时间区间的方式表示）
    now_time = Interval(now_localtime, now_localtime)
    # 在当前时间区间内，在去判断当天是否为交易日
    if now_time in start_time_interval or now_time in end_time_interval:
        return check_trade_date(datetime.today().date(),rainiee_client)
    return False

def check_trade_date(date,rainiee_client):
    return rainiee_client.vip_channel().is_trading_day(date=date.strftime('%Y%m%d'))


def model_to_dict(models):
    from django.forms import model_to_dict
    return_list = []
    for model in models:
        return_list.append(model_to_dict(model, exclude=[]))
    return return_list


import traceback
import boto3
from io import StringIO
from botocore.exceptions import ClientError
from ..lib.utils import *


class BotoClient(object):
    def __init__(self):
        self.ACCESS_KEY = 'AKIAIWPNDYOM6HPUZTBQ'
        self.SECRET_KEY = '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'
        self.s3_client = boto3.client('s3', aws_access_key_id=self.ACCESS_KEY, aws_secret_access_key=self.SECRET_KEY)
        self.s3_resource = boto3.resource('s3', aws_access_key_id=self.ACCESS_KEY, aws_secret_access_key=self.SECRET_KEY)
        self.bucket = 'rainiee'

    def get_client(self):
        return self.s3_client

    def upload_to_aws(self, local_file, s3_file,delete_local_file = False):
        try:
            self.s3_client.upload_file(local_file, self.bucket, s3_file)
            info("Upload Successful")
            if delete_local_file:
                import os
                os.remove(local_file)
        except Exception as e:
            error(traceback.format_exc())
            return False
        return True

    def pandas_to_aws(self, dataframe, file_path, file_name = 'df.csv'):
        try:
            csv_buffer = StringIO()
            dataframe.to_csv(csv_buffer)
            self.s3_resource.Object(self.bucket, str(file_path)+str(file_name)).put(Body=csv_buffer.getvalue())
            info('Successfully uploaded to S3 to ' + str(file_path) + str(file_name))
            return True
        except Exception:
            error(traceback.format_exc())
            return False

    def contains_file(self, file_name):
        try:
            file_list = self.s3_client.list_objects(Bucket = self.bucket)['Contents']
            for file in file_list:
                if file_name in file['Key']:
                    info('Already containing S3 file ' + file_name)
                    return True
            return False
        except Exception:
            error(traceback.format_exc())
            return False

    def check_file_exist(self, file_path, file_name):
        try:
            self.s3_client.head_object(Bucket=self.bucket, Key=file_path+file_name)
        except ClientError as e:
            return int(e.response['Error']['Code']) != 404
        return True



import math
from typing import List
from rainiee_lib.dataframe import pandas_df
class Order(object):
    def construct(self, index : int, symbol : str, order_qty : float, order_price : float, buy_order : bool, ext={}):
        self.index = index
        self.symbol = symbol
        self.order_qty = int(math.floor(order_qty / 100.0)) * 100 #取整批最小100股交易
        self.order_price = order_price
        #默认在固定价格全部成交
        self.commision = 0.0005 #千分之一综合交易成本手续费
        self.buy_order = buy_order
        self.ext = ext

    """
    获取总成交金额（含手续费）
    """
    def get_filled_amount(self) -> float:#假设全部成交
        if(self.buy_order):
            return self.order_qty * self.order_price * (1 + self.commision)
        else:
            return self.order_qty * self.order_price * (1 - self.commision)

class OrderEntries(object):
    def construct(self, orders : List[Order]):
        self.orders = orders

    def get_orders_df(self):
        return pandas_df.get_df_obj_list(self.orders)



from rainiee_lib.dataframe import pandas_df
from rainiee_lib.objects import objects_utils
from rainiee_lib.constants import data_consts
import json
from typing import List
import pandas as pd
class Stock(object):
    def construct(self, json_input):
        self.symbol = json_input['symbol']
        self.symbol_name = json_input['symbol_name']
        self.index = json_input['index']
        self.date = json_input['date']
        self.open_price = json_input['open_price']
        self.high_price = json_input['high_price']
        self.low_price = json_input['low_price']
        self.close_price = json_input['close_price']
        self.pre_close_price = json_input['pre_close_price']
        self.turnover_rate = json_input['turnover_rate']
        self.volume = json_input['volume']
        self.amount = json_input['amount']
        self.total_market_value = json_input['total_market_value']
        self.circulate_market_value = json_input['circulate_market_value']
        self.stock_category = json_input['stock_category']
        self.list_days_index = json_input['list_days_index']
        self.turnover_rate_f = json_input['turnover_rate_f']
        self.volume_ratio = json_input['volume_ratio']
        self.pe = json_input['pe']
        self.pe_ttm = json_input['pe_ttm']
        self.pb = json_input['pb']
        self.ps = json_input['ps']
        self.ps_ttm = json_input['ps_ttm']
        self.dv_ratio = json_input['dv_ratio']
        self.dv_ttm = json_input['dv_ttm']
        self.total_share = json_input['total_share']
        self.float_share = json_input['float_share']
        self.free_share = json_input['free_share']
        self.intra_ret = self.__calc_ret(self.close_price, self.open_price)
        self.inter_ret = self.__calc_ret(self.close_price, self.pre_close_price)
        self.centroid = self.__division(self.high_price + self.low_price, self.pre_close_price) / 2
        self.diff_high = self.__division(self.high_price - self.open_price, self.pre_close_price)
        self.diff_low = self.__division(self.open_price - self.low_price, self.pre_close_price)
        self.volatility = self.__division(self.high_price - self.low_price, self.pre_close_price)

    def __calc_ret(self, end_price, start_price):
        return self.__division(end_price - start_price, start_price)

    def __division(self, numerator, denominator):
        return 0 if denominator == 0 else numerator / denominator

    def to_json(self):
        return self.__dict__

    def from_json(self, json_input):
        self.construct(json_input)
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))


#由于Stock和StockHistory为固定对接数据中台对象接口，不涉及内部逻辑，所以json_output为[{}]形式
class StockHistory(object):
    def construct(self, stock_history: List[Stock]):
        self.stock_history = stock_history

    def get_symbol(self) -> str:
        return self.stock_history[0].symbol

    def get_stock_history_df(self) -> pd.DataFrame:
        stock_history_df = pandas_df.get_df_obj_list(self.stock_history).set_index("date").sort_index()
        cols = [data_consts.OPEN_PRICE, data_consts.HIGH_PRICE, data_consts.LOW_PRICE, data_consts.CLOSE_PRICE, data_consts.PRE_CLOSE_PRICE,
                data_consts.TURNOVER_RATE, data_consts.VOLUME, data_consts.AMOUNT, data_consts.TOTAL_MKT_VAL, data_consts.CIRC_MKT_VAL]
        stock_history_df[cols] = stock_history_df[cols].apply(pd.to_numeric, errors='coerce', axis=1)
        return stock_history_df
    def get_stock_history_df_range(self, start_index:int, end_index:int) -> pd.DataFrame:
        dataframe = self.get_stock_history_df()
        return dataframe[(dataframe['index'] >= start_index) & (dataframe['index'] <= end_index)]
    def get_stock_history_range(self, start_index:int, end_index:int)  -> pd.DataFrame:
        return self.get_stock_history_df_range(start_index, end_index).reset_index().to_dict("records")
    def to_json(self):
        return objects_utils.to_json_object_list(self.stock_history)
    def from_json(self, json_input):
        self.stock_history = objects_utils.from_json_object_list("Stock", json_input)
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))



class StockHistoryFetcher(object):
    def construct(self, symbol_list: list, start_index: int, end_index: int):
        self.__assign(symbol_list, start_index, end_index, False)

    def fetcher(self, stock_history_list : List[StockHistory]):
        self.run_mode = True
        self.stock_history_list = stock_history_list
        self.stock_history_map = {x.get_symbol(): x for x in stock_history_list}

    def to_json(self):
        json_output = {"run_mode":self.run_mode,"symbol_list": self.symbol_list, "start_index":self.start_index, "end_index":self.end_index}
        if self.run_mode:
            json_output.update({"stock_history_list":objects_utils.to_json_object_list(self.stock_history_list)})
        return json_output
    def from_json(self, json_input):
        self.__assign(json_input['symbol_list'],json_input['start_index'],json_input['end_index'], json_input['run_mode'])
        if self.run_mode:
            self.stock_history_list = objects_utils.from_json_object_list("StockHistory",json_input['stock_history_list'])
            self.stock_history_map = {x.get_symbol(): x for x in self.stock_history_list}

    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))

    def __assign(self, symbol_list, start_index, end_index, run_mode):
        self.symbol_list = symbol_list
        self.start_index = int(start_index)
        self.end_index = int(end_index)
        self.run_mode = bool(run_mode)






import json
from typing import List
from rainiee_lib.objects.stocks import stocks
from rainiee_lib.objects import objects_utils
"""
Default:
stock_history_fetcher :: StockHistoryFetcher
funnel_model :: str

Funneled:
funneled_stock_history_list : List[stocks.StockHistory]
exclude_list : List[dict]
"""
class FunnelsRunner(object):
    def construct(self, stock_history_fetcher : stocks.StockHistoryFetcher,
                  funnel_model :str):
        self.__assign(stock_history_fetcher, funnel_model, False)
        self.stock_history_fetcher = stock_history_fetcher
        self.funnel_model = funnel_model

    def run(self, funneled_stock_history_list : List[stocks.StockHistory],
               exclude_list : List[dict]):
        self.run_mode = True
        self.funneled_stock_history_list = funneled_stock_history_list
        self.exclude_list = exclude_list

    def __assign(self, stock_history_fetcher : stocks.StockHistoryFetcher, funnel_model : str, run_mode : bool):
        self.stock_history_fetcher = stock_history_fetcher
        self.funnel_model = funnel_model
        self.run_mode = run_mode

    def to_json(self):
        json_output = {"run_mode":self.run_mode,
                       "funnel_model": self.funnel_model,
                       "stock_history_fetcher":self.stock_history_fetcher.to_json()}
        if self.run_mode:
            json_output.update({"funneled_stock_history_list":objects_utils.to_json_object_list(self.funneled_stock_history_list),
                                "exclude_list":self.exclude_list})
        return json_output
    def from_json(self, json_input):
        self.__assign(objects_utils.from_json_object("StockHistoryFetcher",json_input["stock_history_fetcher"]),
                      json_input["funnel_model"],
                      json_input["run_mode"])
        if(self.run_mode):
            self.funneled_stock_history_list = objects_utils.from_json_object_list("StockHistory", json_input["funneled_stock_history_list"])
            self.exclude_list = json_input["exclude_list"]

    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))




from typing import List

from rainiee_lib.objects.backtesting.backtest_result import PortfolioReturn


def daily_return_percentage(stock):
    if stock['pre_close_price'] == 0:
        return 1
    return stock['close_price'] / stock['pre_close_price']

def intraday_return_percentage(stock):
    return stock['close_price'] / stock['open_price']



def compute_geometric_data(data):
    from functools import reduce
    import operator
    gd = reduce(operator.mul, data) ** (1.0 / len(data))
    if type(gd) == complex:
        return gd.real
    return gd
def compute_profit_mean(data):
    sum = 0
    for it in data:
        sum = sum + it
    return sum / len(data)


def compute_mean(return_list:List[PortfolioReturn]):
    index_map = {}
    data = []
    daily_p_return_list=[]
    i = 0
    for ret in return_list:
        daily_p_return_total = round((ret.daily_p_return_total), 5)
        data.append(daily_p_return_total)
        index_map[daily_p_return_total] = ret.index
        daily_p_return_list.append(ret.daily_p_return)
        i = i + 1
    min_hold_profit = min(data)
    min_hold_profit_index = index_map[min_hold_profit]
    max_hold_profit = max(data)
    max_hold_profit_index = index_map[max_hold_profit]
    return round(compute_geometric_data(daily_p_return_list), 5), round(compute_profit_mean(daily_p_return_list),5), min_hold_profit, min_hold_profit_index, max_hold_profit, max_hold_profit_index


import traceback

from rainiee_data.data_client import DataClient
from rainiee_rest_client.rest_client import RestClient

from ..lib import utils

client_init_dict = {
    'rainiee_data_client':'DataClient(self.username, self.password)',
    'rainiee_compute_client':'RestClient(self.username, self.password).choice_rainiee_compute().login()'
}
client_dict = {}

class ClientFactory(object):
    def __init__(self):
        self.username = 'hhh7788518'
        self.password = 'hu7788518'

    def rainiee_data_client(self):
        return self.__get_client(self.rainiee_data_client.__name__)

    def rainiee_compute_client(self):
        return self.__get_client(self.rainiee_compute_client.__name__)

    def init_all_client(self):
        for key,value in client_init_dict.items():
            self.__init_client(key,value)


    def __get_client(self,key):
        client = client_dict.get(key)
        if client is None:
            self.__init_client(key,client_init_dict[key])
            client = client_dict.get(key)
        return client

    def __init_client(self,key,value):
        try:
            client_dict[key] = eval(value)
        except Exception as e:
            utils.error(traceback.format_exc())




import json

class BaseObject(object):

    def to_json_str(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def from_json(self,json_data:dict):
        self.__dict__ = json_data


from rainiee_lib.dataframe import pandas_df
from rainiee_lib.objects import objects_utils
from rainiee_lib.constants import data_consts
import json
from typing import List
import pandas as pd
class Index(object):
    def construct(self, json_input):
        self.symbol = json_input['symbol']
        self.symbol_name = json_input['symbol_name']
        self.index = json_input['index']
        self.date = json_input['date']
        self.open_price = float(json_input['open_price'])
        self.high_price = float(json_input['high_price'])
        self.low_price = float(json_input['low_price'])
        self.close_price = float(json_input['close_price'])
        self.pre_close_price = float(json_input['pre_close_price'])
        self.volume = float(json_input['volume'])
        self.amount = float(json_input['amount'])
        self.intra_ret = self.__calc_ret(self.close_price, self.open_price)
        self.inter_ret = self.__calc_ret(self.close_price, self.pre_close_price)
        self.centroid = (self.high_price + self.low_price) / self.pre_close_price / 2
        self.diff_high = (self.high_price - self.open_price) / self.pre_close_price
        self.diff_low = (self.open_price - self.low_price) / self.pre_close_price
        self.volatility = (self.high_price - self.low_price) / self.pre_close_price

    def __calc_ret(self, end_price, start_price):
        if (start_price != 0):
            return (end_price - start_price) / start_price
        else:
            return 0
    def to_json(self):
        return self.__dict__
    def from_json(self, json_input):
        self.construct(json_input)
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))

class IndexHistory(object):
    def construct(self, index_history: List[Index]):
        self.index_history = index_history

    def get_symbol(self) -> str:
        return self.index_history[0].symbol

    def get_index_history_df(self) -> pd.DataFrame:
        index_history_df = pandas_df.get_df_obj_list(self.index_history).set_index("date").sort_index()
        cols = [data_consts.OPEN_PRICE, data_consts.HIGH_PRICE, data_consts.LOW_PRICE, data_consts.CLOSE_PRICE, data_consts.PRE_CLOSE_PRICE]
        index_history_df[cols] = index_history_df[cols].apply(pd.to_numeric, errors='coerce', axis=1)
        return index_history_df
    def get_index_history_df_range(self, start_index:int, end_index:int) -> pd.DataFrame:
        dataframe = self.get_index_history_df()
        return dataframe[(dataframe['index'] >= start_index) & (dataframe['index'] <= end_index)]
    def get_index_history_range(self, start_index:int, end_index:int)  -> pd.DataFrame:
        return self.get_index_history_df_range(start_index, end_index).reset_index().to_dict("records")
    def to_json(self):
        return objects_utils.to_json_object_list(self.index_history)
    def from_json(self, json_input):
        self.index_history = objects_utils.from_json_object_list("Index", json_input)
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))



class IndexHistoryFetcher(object):
    def construct(self, symbol_list: list, start_index: int, end_index: int):
        self.__assign(symbol_list, start_index, end_index, False)

    def fetcher(self, index_history_list : List[IndexHistory]):
        self.run_mode = True
        self.index_history_list = index_history_list
        self.index_history_map = {x.get_symbol(): x for x in index_history_list}

    def to_json(self):
        json_output = {"run_mode":self.run_mode,"symbol_list": self.symbol_list, "start_index":self.start_index, "end_index":self.end_index}
        if self.run_mode:
            json_output.update({"index_history_list":objects_utils.to_json_object_list(self.index_history_list)})
        return json_output
    def from_json(self, json_input):
        self.__assign(json_input['symbol_list'],json_input['start_index'],json_input['end_index'], json_input['run_mode'])
        if self.run_mode:
            self.index_history_list = objects_utils.from_json_object_list("IndexHistory",json_input['index_history_list'])
            self.index_history_map = {x.get_symbol(): x for x in self.index_history_list}
    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))

    def __assign(self, symbol_list, start_index, end_index, run_mode):
        self.symbol_list = symbol_list
        self.start_index = int(start_index)
        self.end_index = int(end_index)
        self.run_mode = bool(run_mode)






from typing import List
from rainiee_lib.constants import data_consts
from rainiee_lib.objects.stocks import stocks
from rainiee_lib.objects import objects_utils
from rainiee_lib.dataframe import pandas_df
from rainiee_lib.lib import utils
import pandas as pd
import json
class RetPredictions(object):
    def construct(self, stock_history_list : List[stocks.StockHistory],
                  ret_predict_model : str, param = {}):
        self.__assign(stock_history_list, ret_predict_model, param, False)

    def get_returns_df(self):
        returns_df = []
        utils.info('STARTED to enumerate symbol list')
        for stock_history in self.stock_history_list:
            stock_history_df = stock_history.get_stock_history_df()
            symbol = stock_history_df[data_consts.SYMBOL][0]
            stock_history_df[symbol] = stock_history_df['inter_ret']
            hist_df = stock_history_df[[symbol]]
            returns_df = self.__merge_final_df(returns_df, hist_df, symbol)
        utils.info('COMPLETED to enumerate symbol list')
        return returns_df

    def predict(self, mean_returns, cov_matrix):
        self.run_mode = True
        self.mean_returns = mean_returns
        self.cov_matrix = cov_matrix

    def to_json(self):
        json_output = {"run_mode":self.run_mode,
                       "ret_predict_model":self.ret_predict_model,
                       "stock_history_list" :objects_utils.to_json_object_list(self.stock_history_list),
                       "param": self.param}
        if self.run_mode:
            json_output.update({
                "mean_returns":pandas_df.json_write_dataframe(self.mean_returns),
                "cov_matrix":pandas_df.json_write_dataframe(self.cov_matrix)
            })
        return json_output

    def from_json(self, json_input):
        self.__assign(
            objects_utils.from_json_object_list("StockHistory", json_input["stock_history_list"]),
            json_input["ret_predict_model"],
            json_input["param"],
            json_input["run_mode"]
        )
        if self.run_mode:
            self.mean_returns = pandas_df.json_read_dataframe(json_input['mean_returns'])
            self.cov_matrix = pandas_df.json_read_dataframe(json_input['cov_matrix'])

    def to_json_str(self):
        return json.dumps(self.to_json())
    def from_json_str(self, json_str):
        self.from_json(json.loads(json_str))

    def __merge_final_df(self, final_df, hist_df, symbol):
        if len(final_df) == 0:
            final_df = hist_df
        else:
            try:
                final_df = pd.merge(final_df, hist_df, on='date', how='inner')
            except Exception as e:
                utils.info('Excluding symbol [{}] due to exception on join'.format(symbol))
        return final_df

    def __assign(self, stock_history_list : List[stocks.StockHistory],
                 ret_predict_model: str,
                 param : dict,
                 run_mode: bool):
        self.run_mode = run_mode
        self.stock_history_list = stock_history_list
        self.ret_predict_model = ret_predict_model
        self.param = param



PORTF_FRAC = 'frac'
PORTF_FRAC_MAX = 'frac_max'

PORTF_MVO_MRG = 'mean_return_gt'
PORTF_MVO_MRM = 'mean_return_multiplier'
PORTF_MVO_RM = 'return_min'
PORTF_MVO_MIQP = 'portfolio_miqp'

PORTF_MAX_SHARP_RF = 'risk_free_rate'
PORTF_MAX_SHARP_SLSQP = 'SLSQP'